;; wwwpic2 is a tcp/ip stack and www server in a PIC 16F84
;;
;; This program is free software;  you can redistribute it and/or
;; modify it under the terms of the GNU General Public License
;; as published by the Free Software Foundation;  either version 2
;; of the License, or (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY;  without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program;  if not, write to the Free Software
;; Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

;; To contact V Sanders email
;; vince@kyllikki.fluff.org

;; The main web site for this program is
;; http:	//www.kyllikki.fluff.org/hardware/wwwpic2/
;;
;; This program is based upon wwwpic by Denis Petrov <zhengxi@operamail.com>
;; From and original idea by Shrikumar H. <shri@cs.umass.edu>
;;
;; See the website and acompanying README for more details

;; Version 0.11
	
; define PIC device type
;
        device  pic16f84

;
; define config fuses
;
        config  CP=off,WDT=off,PWRT=off,OSC=hs

;
; include PIC register definitions, and some macros
;
        include "pic16c84.h" ; this include also works with 16F84
        include "picmac.h"

; local include
	include "wwwpic2.h"


;;========================================================================
;; General settings
;;========================================================================
QUARTZ_FREQ             equ  10000000
RS_BAUD                 equ    19200
RS_DATA                 equ        8
RS_TICKS                equ  (((QUARTZ_FREQ/4)/RS_BAUD)+1); add one because
				 ; of fractions see _delay_x_9 for details

ADDR_MYSELF_0           equ      192      ; 192.168.2.2 - private net addr
ADDR_MYSELF_1           equ      168
ADDR_MYSELF_2           equ        2
ADDR_MYSELF_3           equ        2

TCP_PORT_HTTP           equ       80      ; 1..255

TCP_RCV_WINDOW          equ      128;!sic!; [non-volatile]
TCP_SND_WINDOW          equ      128      ; [non-volatile]

FS_UNIT_SIZE            equ      128      ; [non-volatile]
FS_ROOT_DIR_ENTRIES     equ        8      ; min 2: INDEX and ERROR.

FS_USE_INTERNAL_ROM     equ        1
FS_INTERNAL_ROM_UNITS   equ        4      ; number of 128-byte pages
FS_USE_INTERNAL_EEPROM  equ        0      ;

FS_USE_EXTERNAL_ROM     equ        0


;;========================================================================
;; DATA RAM mapping
;;========================================================================
org 0x0c,reg

TCPR_RPORT_H            ds  1; remote TCP port
TCPR_RPORT_L            ds  1;
TCPR_LPORT_H            ds  1; local TCP port, TCPR_LPORT_H==0
TCPR_LPORT_L            ds  1;
TCPR_SEQ_L3             ds  1;
TCPR_SEQ_L2             ds  1;
TCPR_SEQ_L1             ds  1;
TCPR_SEQ_L0             ds  1;
TCPR_ACK_L3             ds  1;
TCPR_ACK_L2             ds  1;
TCPR_ACK_L1             ds  1;
TCPR_ACK_L0             ds  1;
TCPR_HLEN               ds  1;
TCPR_FLAG               ds  1;

TEMP_0                  ds  1;
TEMP_1                  ds  1;
TEMP_2                  ds  1;

HTTP_UNITIDX            ds  1;
HTTP_FILESIZE_H         ds  1;
HTTP_FILESIZE_L         ds  1;
HTTP_NAME_0             ds  1;
HTTP_NAME_1             ds  1;
HTTP_NAME_2             ds  1;

; socket data, it's valid if tcp_online==1
TCPR_POS                ds  1		    ; iterator 0 -->] 80..FF

TCPS_ACK                ds  1;
TCPS_SEQ                ds  1;
TCPS_LEN                ds  1		    ; for retransmit
TCPS_FLAG               ds  1		    ; for retransmit

RX_BYTE                 ds  1		    ; [soft uart]
TX_BYTE                 ds  1		    ; [soft uart]

IP_POS                  ds  1		    ; iterator
IP_HLEN                 ds  1		    ; IP header len (incoming only)
IP_LEN                  ds  1		    ; IP packet len
IP_SUM_0                ds  1		    ; TCP/IP chksum
IP_SUM_1                ds  1		    ; --//-----//-- low octet
IP_ADR_0                ds  1		    ; remote IP address
IP_ADR_1                ds  1		    ; --//--
IP_ADR_2                ds  1		    ; --//--
IP_ADR_3                ds  1		    ; --//--

B_HTTP_STATE_0		equ 5		    ; in TCPS_FLAG
B_HTTP_STATE_1		equ 6		    ; in TCPS_FLAG
B_TCP_ONLINE		equ 7		    ; in TCPS_FLAG socket data valid
B_RS232_RxD		equ 0		    ; bit 0 on PORTB (RB0)
B_RS232_TxD		equ 3		    ; bit 3 on PORTA (RA3)
B_RS232_CTS		equ 2	            ; bit 2 on PORTA (RA2)
; discard data flag - when set slip recive routines simply discards data
B_SLIP_OUTPKT		equ 7	            ;
; escape flag set when previous char was a slip escape char
B_SLIP_ESCAPE		equ 6	            ;

;;========================================================================
;; Start Code
;;========================================================================
org 0,code
_start:
        clrf        INTCON                  ; disables all interups clears GIE
					    ; and all pending interupts
        movlw       01001111b               ; (depricated) sets options
        option			            ; PORTB pulups on
					    ; Rising edge trigger on RB0
					    ; TMR0 clock source internal
					    ; TMR0 low to high edge
					    ; prescaler asigned to WDT
					    ; prescaler set to 1:	128
        goto        _init		    ; goto start routine

;;========================================================================
;; Interrupt Code
;;========================================================================
org 4,code
;; Called from main interupt vector as RBO edge is only active interupt source
;;  it came from there!
	;; Soft UART - simplex :(
	;; first recive a byte at the pre-set baud rate 8Data No Parity X stop
_rx_byte:                                   ; 5 ; well 3-4 (interupt latency)
        call        _delay4                 ; 4 ; delays 4 cycles total
        movlw       8                       ; 1 ; 8data No parity x stop
_rx_bit:
        bcf         PORTA,B_RS232_CTS       ; 1 ; drop CTS to stop any more
	                                    ;       bytes being sent while were
		                            ;	    dealing with this one
        call        _delay_x_9              ; 2 ; RS_TICKS-9
        bcf         STATUS,C                ; 1 ;
        skip1       PORTB,B_RS232_RxD       ; 1 ;
        bsf         STATUS,C                ; 1 ; C:=RS232_RxD
        rrf         RX_BYTE,F               ; 1 ;
        addlw       -1                      ; 1 ;
        jnz         _rx_bit                 ; 3 ; round for next bit

;; RX SLIP code
	;; RX_BYTE register contains recived byte
        movf        RX_BYTE,W               ; get recived byte into w
        xorlw       0C0h                    ;
        jnz         _not_slip_end_char      ; check for SLIP END char
	; recivedSLIP END
        bcf         IP_HLEN,B_SLIP_OUTPKT   ; clear discard data flag
        clrf        IP_POS                  ; clear IP_POS
        clrf        IP_SUM_0                ; clear ip checksup low byte
        clrf        IP_SUM_1                ; clear ip checksup high byte
        jump        _rx                     ; return to busy loop

_not_slip_end_char:
        brclr       IP_HLEN,B_SLIP_OUTPKT,_mode_slip ; is discard data flg set

_mode_terminal:
;; terminal mode - discarding data any 0xD char codes reply with ok
        xorlw       (00Dh ^ 0C0h)           ; Check for a 0xD char
        jnz         _rx                     ; Not a 0xD return to busy loop
        TXLW        'O'                     ; I think this is for slip
        TXLW        'K'                     ;   server sw that trys to issue
        TXLW        0Dh                     ;   AT commands at startup
        jump        _rx                     ; return to busy loop

_mode_slip:
;; slip mode - not discarding data
        brclr       IP_HLEN,B_SLIP_ESCAPE,_no_slip_escape; check escape flag
;; previous char was slip escape char
;; PROTOCOL BUG - assumes its 0xDC or 0xDD - should pass thru if neither
;;   see RFC 1055
	xorlw       (0DCh ^ 0C0h)           ; check for 0xDC
        movlw       0DBh                    ; Escaping in SLIP:
        skipnz                              ; 0xDB,0xDC --> 0xC0
        movlw       0C0h                    ; 0xDB,0xDD --> 0xDB
        movwf       RX_BYTE                 ; RX_BYTE=(RX_BYTE==0xDC?0xC0:0xDB)
;       jump        _end_slip_escape         ; ok ,w either 0xdc or 0xdd so
		                            ;   later xor fails

_no_slip_escape:                            ;
; previous char was not an escape character - check this one
        bsf         IP_HLEN,B_SLIP_ESCAPE   ; set escape flag to true
        xorlw       (0DBh ^ 0C0h)           ; check for SLIP_escape
        jz          _rx                     ; ret to busy loop with esc flg set

_end_slip_escape:
        bcf         IP_HLEN,B_SLIP_ESCAPE   ; clear escape flag


;; byte recived and slip de-escaped and framed
        movf        RX_BYTE,W	            ; retrieve our original byte
        call        _checksum_byte          ; calculates the ip checksum
	                                    ;   as we go

        movf        IP_POS,W                ; rus com
        addlw       -20                     ;(rus com) SLIP.END (rus com)
        jc          _xxxx                   ; IP (rus com)
        rlf         IP_POS,W                ;
        clrf        PCLATH                  ; (rus)
        addwf       PCL,F                   ;

        movf        RX_BYTE,W               ; IPhdr+ 0  Version & IHL
        movwf       IP_HLEN                 ;
        rlf         IP_HLEN,F               ; IPhdr+ 1  Type of Service
        jump        _01_                    ;
        movlw       0                       ; IPhdr+ 2  Total Length (hi)
        jump        _rx_byte_must_equal_w   ;
        movf        RX_BYTE,W               ; IPhdr+ 3  Total Length (lo)
        movwf       IP_LEN                  ;
        jump        _finish                 ; IPhdr+ 4  Identification (hi)
_01_:   bcf         IP_HLEN,7               ;
        jump        _finish                 ; IPhdr+ 5  Identification (lo)
        nop                                 ;
        jump        _finish                 ; IPhdr+ 6  Flags & Fragment Offset (hi)
        nop                                 ;
        movlw       0                       ; IPhdr+ 7  Fragment Offset (lo)
        jump        _rx_byte_must_equal_w   ;
        jump        _finish                 ; IPhdr+ 8  Time to Live
        nop                                 ;
        movlw       6                       ; IPhdr+ 9  Protocol
        jump        _rx_byte_must_equal_w   ;
        jump        _finish                 ; IPhdr+10  Header Checksum (hi)
        nop                                 ;
        jump        _finish                 ; IPhdr+11  Header Checksum (lo)
        nop                                 ;
        movlw       IP_ADR_0                ; IPhdr+12  Source Address
        jump        _save_rx_byte           ;
        movlw       IP_ADR_1                ; IPhdr+13
        jump        _save_rx_byte           ;
        movlw       IP_ADR_2                ; IPhdr+14
        jump        _save_rx_byte           ;
        movlw       IP_ADR_3                ; IPhdr+15
        jump        _save_rx_byte           ;
        movlw       ADDR_MYSELF_0           ; IPhdr+16  Destination Address
        jump        _rx_byte_must_equal_w   ;
        movlw       ADDR_MYSELF_1           ; IPhdr+17
        jump        _rx_byte_must_equal_w   ;
        movlw       ADDR_MYSELF_2           ; IPhdr+18
        jump        _rx_byte_must_equal_w   ;
        movlw       ADDR_MYSELF_3           ; IPhdr+19
_rx_byte_must_equal_w:
        xorwf       RX_BYTE,W               ;
        jnz         _error                  ;

_xxxx:
        movf        IP_HLEN,W               ; IP_POS-IP_HLEN - (rus)
        subwf       IP_POS,W                ; (rus) IP (rus)
        jc          _post_ip                ; ... (rus)?

_no_post_ip:
        xorlw       0FFh                    ; (rus) IP (rus)
        jnz         _finish                 ; IP_SUM (rus)
        movf        IP_SUM_0,W              ;
        iorwf       IP_SUM_1,W              ;
        jnz         _error                  ; (rus)
_finish:
        incf        IP_POS,F                ; (rus)
        jump        _rx                     ;

_post_ip:
        addlw       -14                     ; (rus) TCP (rus)
        jnc         _save_as_tcp_header     ;
        jnz         _no_correct_tcp_hlen    ;
        rrf         TCPR_HLEN,F             ; if(IP_POS-IP_HLEN==14)
        rrf         TCPR_HLEN,F             ; {
        bcf         TCPR_HLEN,7             ;    TCPR_HLEN=(TCPR_HLEN<<2)&0x3F;
        bcf         TCPR_HLEN,6             ; }
_no_correct_tcp_hlen:
        subwf       TCPR_HLEN,W             ; (rus) TCP (rus)
        addlw       -15                     ; (rus) (W<14) - (rus)
        jc          _no_save_rx_byte        ; (rus)
                                            ;  GET /xxx (rus) (rus) telnet-(rus)
                                            ; (rus)
                                            ; (rus) www-browser.
        subwf       TCPR_POS,W              ; (rus) "GET /.."
        addlw       (-1-(5+3))              ; W=TCPR_POS-(TCPR_HLEN-(IP_POS-IP_HLEN-14))
        addlw       3                       ; (5.7)->(FD.FF) (5.7)->(0.2)
        jnc         _no_save_rx_byte
if ((HTTP_NAME_0 - (TCPR_RPORT_H + 14)) != 0)
        addlw       (HTTP_NAME_0 - (TCPR_RPORT_H + 14))
endif
_save_as_tcp_header:
        addlw       TCPR_RPORT_H+14
_save_rx_byte:
        movwf       FSR                     ; [W]:=RX_BYTE - (rus)
        movf        RX_BYTE,W               ;
        movwf       IND0                    ;

_no_save_rx_byte:
        incf        IP_POS,W                ; (rus)
        subwf       IP_LEN,W                ; IP (rus)...
        jnz         _finish                 ;

        movf        IP_HLEN,W               ; IP_LEN-=IP_HLEN;
        subwf       IP_LEN,F                ;
        call        _chksum_pseudoheader    ; (rus)
        movf        IP_SUM_0,W              ; TCP (rus)
        iorwf       IP_SUM_1,W              ;
        skipnz                              ;
        call        _process_tcp            ;
        jump        _error                  ;



;===========================================================================
; TX routines
;===========================================================================
;_txhex: movwf       FSR
;        swapf       FSR,W
;        call        _txhex0
;        movf        FSR,W
;_txhex0:andlw       00001111b
;        addlw       -10
;        skipnc
;        addlw       7
;        addlw       58
;        jump        _tx
;;------------------------------------------------
_slitx_csum:
        movf        IP_SUM_0,W              ; (rus)
        call        _slitx                  ; (rus)
        movf        IP_SUM_1,W              ;
        jump        _slitx
;;------------------------------------------------
_slitx4:
        movwf       FSR                     ; (rus) (W[0]..W[3])
        call        _slitx2                 ; (rus)
_slitx2:
        movf        IND0,W                  ; (rus) (FSR[0]..FSR[1])
        call        _checksum_hi
        call        _slitx1
        movf        IND0,W
        call        _checksum_lo
_slitx1:
        movf        IND0,W                  ; (rus) (*FSR++)
        incf        FSR,F                   ;
        jump        _slitx
;;------------------------------------------------
_tx_0_0:
        call        _tx_0                   ; (rus)
_tx_0:
        movlw       000h                    ; (rus)
_slitx:
        xorlw       0C0h                    ; (rus) SLIP
        jnz         _ne_C0                  ; (rus)
        TXLW        0DBh                    ;   0xC0 --> 0xDB,0xDC
        movlw       0DCh                    ;   0xDB --> 0xDB,0xDD
        jump        _tx                     ;
_ne_C0:
        xorlw       (0DBh ^ 0C0h)           ; <0DBh xor 0C0h>
        jnz         _tx11                   ;
        call        _tx11                   ;
        movlw       (0DDh ^ 0DBh)           ; <0DDh xor 0DBh>
_tx11:                                      ;
        xorlw       0DBh                    ;
_tx:
        bsf         PORTA,B_RS232_TxD       ; rus com
        movwf       TX_BYTE                 ; IN:  W
        movlw       ((-(RS_DATA+1))<< 4) & 0xff ; <(-(RS_DATA+1))shl 4> ; OUT: TX_BYTE:=W
_tx_loop_c:
        bsf         STATUS,C                ; 1
_tx_loop:
        call        _delay_x_9
        rrf         TX_BYTE,F               ; 1 ; 1
        jnc         _c_is_clear             ; 2 ; 3
        addlw       10h                     ; 1
        bcf         PORTA,B_RS232_TxD       ; 1
        jnz         _tx_loop_c              ; 3

;; Delays for RS_TICKS less a fixed value (9 is the current value)
;; currently depends on 6Mhz clock! - possible bug in that it delays for
;; 70 cycles (total 79) where RS_TICKS is 78 (see next bit for reason)
;; For 10Mhz operation	:
;; RS_TICKS will be 130.2 hence we either creep back by 0.2 or add an extra
;;   cycle and creep forward 0.8.
;; Given how close Denis starts to the bits leading edge, I suspect that he
;;   delibarately added the extra cycle at 6Mhz to cause forward creep rather
;;   than back.
;; I shall do likewise and cause a delay for 122 cycles after 8 bits this only
;;   puts us 6/1/2 cycles futher along along (8%) anyway.
;; I wont do this here tho! i shall add it to RS_TICKS at the top

_delay_x_9:
;; Smallest delay possible with this routine is 19 cycles (38400 @ 4Mhz clock
;; is possible using 26 cycles so i dont think its an issue )
;; Uses TEMP_0 and TEMP_1 and 7 instrctions
;; Delay comprises of
;;   9 - static value
;;   2 - call to get here
;;   6 - routine overhead
;;   (n*3) - where n is number of loops required
;;  -1 - (-1 coz last time round  loop only executes 2 cycles not 3)
;;   r - where r is (n mod 3) to get number of extras leftover
;;
;; We need to calculate sumber of times round loop which is RS_TICKS less the
;;   static values (9+2+6-1)=16

WAIT_CYCLES set (RS_TICKS - 16)
WAIT_LOOPS set (WAIT_CYCLES/3)
if (WAIT_LOOPS < 1)
	error
endif
	movwf       TEMP_1	; (1) store w to temp_1
	movlw	    WAIT_LOOPS	; (1) get number of loops to do
WAIT_CYCLES set (WAIT_CYCLES - (WAIT_LOOPS*3)); calc number of spare cycles
	movwf	    TEMP_0	; (1) store literal in TEMP_0
	movf	    TEMP_1,W	; (1) recover w from TEMP_1
_tx_delay:
        decfsz      TEMP_0,F    ; (1) decrement and jump if zero
        goto        _tx_delay	; (2(1 if skipped)) go back round loop
				; we are one cycle down coz the goto got missed
				;   have to account for it with a nop
if (WAIT_CYCLES == 1)
	nop			; (1)
endif
if (WAIT_CYCLES == 2)
	goto _delay_x_9_next2	; (2)
_delay_x_9_next2:
endif
_delay4:
        return			; (2)

_c_is_clear:
        bsf         PORTA,B_RS232_TxD       ; 1
        addlw       10h                     ; 1
        bcf         STATUS,C                ; 1  =nop
        jump        _tx_loop                ; 2



;;========================================================================
;; Primary initialisation - na d main loop
;;========================================================================
_init:
                                            ; RS232_TxD:=0, RS232_CTS:=0
        clrf        PORTB                   ; Initialise PORTB to off
        clrf        PORTA		    ; Initialise PORTA to off
        movlw       00000000b               ; porta, ___ooooo
        tris        5			    ; (depreciated) set PORTA data
					    ; direction to all output
        movlw       00000001b               ; portb, oooooooi
        tris        6			    ; (depreciated) set PORTB to
					    ; output except for RB0

        bcf         TCPS_FLAG,B_TCP_ONLINE  ; clear tcp_online flag
;; return here upon recipt of error
_error:
        clrf        IP_HLEN                 ; clear IP_HLEN reg
        bsf         IP_HLEN,B_SLIP_OUTPKT   ; set discard data flag
;; return here to continue to wait for next byte
_rx:
        bsf         PORTA,B_RS232_CTS       ; turn on CTS
        movlw       10010000b               ; enable GIE and INTE
        movwf       INTCON	            ;
_busy_loop:
; The interupt routine doesnt use RETIE it simply jumps back to either of the
;   _rx or _error entry points
        jump        _busy_loop	            ; infinite busy loop

;========================================================================
; (rus) TCP/IP (rus)
;========================================================================
_just_ack:
_tcp_send_empty_10:
        bcf         TCPS_FLAG,0             ; TCPS_FLAG=xxxxxxx0
        bcf         TCPS_FLAG,1             ; TCPS_FLAG=xxxxxx0x
_tcp_send_empty:
        clrf        TCPS_LEN
_tcp_send:
        call        _tx_slip_end
        SUM0        (-(4500h+4000h+8000h))  ; (rus) IP (rus)
        movwf       IP_SUM_0
        SUM1        (-(4500h+4000h+8000h))
        movwf       IP_SUM_1
        TXLW        045h
        call        _tx_0_0
        movf        TCPS_LEN,W
        addlw       14h+14h
        movwf       IP_LEN                  ; IP_LEN = TCPS_LEN+14h+14h
        call        _slitx
        call        _chksum_pseudoheader
        call        _tx_0_0
        TXLW        040h
        call        _tx_0
        TXLW        080h
        TXLW        006h
        call        _slitx_csum
        SLITXLW     ADDR_MYSELF_0
        SLITXLW     ADDR_MYSELF_1
        SLITXLW     ADDR_MYSELF_2
        SLITXLW     ADDR_MYSELF_3
        movlw       IP_ADR_0
        call        _slitx4

        movf        TCPR_SEQ_L0,W           ; (rus) SEQ (rus)
        subwf       TCPS_ACK,W              ; (rus) TCP (rus)
        jc          _no_correct_hi_seq      ;
        incf        TCPR_SEQ_L1,F           ;
        skipnz                              ;
        incf        TCPR_SEQ_L2,F
        skipnz
        incf        TCPR_SEQ_L3,F
_no_correct_hi_seq:
        addwf       TCPR_SEQ_L0,F           ; TCPR_SEQ_L0+=TCPS_ACK-TCPR_SEQ_L0
        SUM0        (-(5000h+TCP_RCV_WINDOW-0014h))
        movwf       IP_SUM_0                ; (rus) TCP (rus)
        SUM1        (-(5000h+TCP_RCV_WINDOW-0014h))
        movwf       IP_SUM_1
        call        _chksum_pseudoheader
        movlw       TCPR_LPORT_H
        movwf       FSR
        call        _slitx2
        movlw       TCPR_RPORT_H
        movwf       FSR
        call        _slitx2
        movlw       TCPR_ACK_L3
        call        _slitx4
        movlw       TCPR_SEQ_L3
        call        _slitx4
        TXLW        050h
        movf        TCPS_FLAG,W
        andlw       00010011b               ; 0x10,0x11,0x12
        call        _tx
        movf        TX_BYTE,W
        call        _checksum_lo
        call        _tx_0
        TXLW        TCP_RCV_WINDOW

        clrf        IP_POS
_loop_top0:
        movf        TCPS_LEN,W              ; (rus)
        xorwf       IP_POS,W                ; (rus) TCP (rus)
        jz          _loop_break0
        call        _file_read
        call        _checksum_byte
        incf        IP_POS,F
        jump        _loop_top0
_loop_break0:
        call        _slitx_csum
        call        _tx_0_0

        clrf        IP_POS                  ; (rus) TCP (rus)
_loop_top1:
        movf        TCPS_LEN,W
        xorwf       IP_POS,W
        jz          _loop_break1
        call        _file_read
        call        _slitx
        incf        IP_POS,F
        jump        _loop_top1
_loop_break1:

_tx_slip_end:
        movlw       0C0h                    ; (rus) SLIP.END
        jump        _tx


;========================================================================
; (rus) TCP/IP (rus)
;========================================================================
_chksum_pseudoheader:
        movf        IP_LEN,W
        call        _checksum_lo
        movf        IP_ADR_0,W
        call        _checksum_hi
        movf        IP_ADR_1,W
        call        _checksum_lo
        movf        IP_ADR_2,W
        call        _checksum_hi
        movf        IP_ADR_3,W
        call        _checksum_lo
        SUM0        (6+((ADDR_MYSELF_0+ADDR_MYSELF_2)<<8)+(ADDR_MYSELF_1+ADDR_MYSELF_3))
        call        _checksum_hi
        SUM1        (6+((ADDR_MYSELF_0+ADDR_MYSELF_2)<<8)+(ADDR_MYSELF_1+ADDR_MYSELF_3))
        jump        _checksum_lo
;------------------------------------------
_checksum_byte:
        brset       IP_POS,0,_checksum_lo
_checksum_hi:
        subwf       IP_SUM_0,F              ; (rus)
        movlw       1
        skipc
_checksum_lo:
        subwf       IP_SUM_1,F              ; (rus)
        movlw       1
        skipc
        subwf       IP_SUM_0,F
        skipc
        subwf       IP_SUM_1,F
        return


;========================================================================
; UDP handler
;========================================================================
_process_udp:
       ; ... to do ...


;========================================================================
; TCP handler
;========================================================================
_process_tcp:
        movf        TCPR_LPORT_L,W          ; (rus)
        xorlw       TCP_PORT_HTTP
        iorwf       TCPR_LPORT_H,W
        skipz
        return

        skip0       TCPR_FLAG,2             ; RST
        bcf         TCPS_FLAG,B_TCP_ONLINE
        brset       TCPR_FLAG,0,_tcp_fin    ; FIN
        brset       TCPR_FLAG,1,_tcp_syn    ; SYN
        skip1       TCPR_FLAG,4             ; ACK ????
        return
;_tcp_ack:
        movf        TCPS_SEQ,W
        subwf       TCPR_ACK_L0,W
        subwf       TCPS_LEN,W              ; TCPS_LEN-TCPR_ACK_L0+TCPS_SEQ
        jz          _my_transmit_is_acked
        subwf       TCPS_LEN,W              ; -TCPR_ACK_L0+TCPS_SEQ
        jz          _tcp_send               ; (rus)... RETRANSMIT !!!!
        return                              ; (rus) TCPR_ACK_L0 = (rus)
_my_transmit_is_acked:
        movf        TCPR_ACK_L0,W           ; tcps_seq=LO(tcpr.ack);
        movwf       TCPS_SEQ                ;
        movf        TCPR_SEQ_L0,W           ; if(tcpr.seq==tcps_ack1)
        xorwf       TCPS_ACK,W              ;
        jnz         _tcp_send               ; (rus)... RETRANSMIT !!!!
;_new_data_was_rvcd:
        movf        TCPR_HLEN,W
        subwf       IP_LEN,W
        addwf       TCPS_ACK,F              ; tcps_ack1 += ip_len-tcpr.hlen;

        addwf       TCPR_POS,F
        skipnc
        bsf         TCPR_POS,7              ; 0-->](128..255)


        ;;****** HTTP server (rus) *******************
        movlw       00010001b               ; http_state    = *
                                            ; b_tcp_online  = 0
                                            ; tcps_flag     = FIN+ACK
        brset       TCPS_FLAG,B_HTTP_STATE_1,_set_flag_and_send
        brset       TCPS_FLAG,B_HTTP_STATE_0,_http_state_1
_http_state_0:

        movf        TCPR_POS,W              ; if(tcpr_pos<*) goto _tcp_send_empty_10
        addlw       -8 ; ***                ; strlen("GET / HTTP/0.9\r\n\r\n")=18
        jnc         _tcp_send_empty_10

                                            ; ??? todo: file system functions
                                            ;=set http_unitidx
                                            ;=set http_filesize_h
                                            ;=set http_filesize_l
if FS_USE_INTERNAL_ROM
                                            ;------------------------
                                            ; CGI
                                            ;------------------------
        movf        HTTP_NAME_0,W           ;
        sublw       'a'                     ;
        jz          _cgi_a                  ; HTTP_NAME_0 == 'a'
        sublw       ('a'-'0')               ;<'a'-'0'>               ;
        addlw       -10                     ;
        jnc         _cgi_digit              ; HTTP_NAME_0 in ['0'..'9']
        movlw       3                       ;
        movwf       HTTP_UNITIDX            ; HTTP_UNITIDX = 3
        clrf        HTTP_FILESIZE_H         ;
        movlw       06Fh                    ; HTTP_FILESIZE = 0x6F
        jump        _cgi_done               ;
_cgi_digit:
        rlf         HTTP_NAME_1,F           ; status.c=0
        movf        HTTP_NAME_1,W           ;
        rlf         HTTP_NAME_1,F           ;
        rlf         HTTP_NAME_1,F           ;
        addwf       HTTP_NAME_1,W           ;
        addwf       HTTP_NAME_2,W           ; W := 10*HTTP_NAME_0+HTTP_NAME_1
        addlw       -16                     ; W := 10*(HTTP_NAME_0-'0')+(HTTP_NAME_1-'0')
        skip0       HTTP_NAME_0,0
        addlw       100
        skip0       HTTP_NAME_0,1
        addlw       200
        movwf       PORTB                   ;

;       andlw       11110b
;       movwf       PORTB                   ;
_cgi_a:
        clrf        HTTP_UNITIDX            ; HTTP_UNITIDX = 0
        movlw       1                       ;
        movwf       HTTP_FILESIZE_H         ; HTTP_FILESIZE = 0x180
        movlw       080h                    ;
_cgi_done:
        movwf       HTTP_FILESIZE_L         ;

endif
if FS_USE_EXTERNAL_ROM
        ...todo...
endif
        bsf         TCPS_FLAG,B_HTTP_STATE_0; http_state=1
        jump        _send_portion_of_file

_http_state_1:
        incf        HTTP_UNITIDX,F          ; http_unitidx++
        movf        TCPS_LEN,W              ; http_filesize -= tcps_len;
        subwf       HTTP_FILESIZE_L,F
        skipc
        decf        HTTP_FILESIZE_H,F
_send_portion_of_file:
        movf        HTTP_FILESIZE_H,F
        jnz         _transmit_80
        movf        HTTP_FILESIZE_L,W
        addlw       7Fh
        jc          _transmit_80
        bsf         TCPS_FLAG,B_HTTP_STATE_1; http_state=3
        movf        HTTP_FILESIZE_L,W       ; <== TRANSMIT less than TCP_SND_WINDOW bytes
        jump        _transmit_not_80
_transmit_80:
        movlw       TCP_SND_WINDOW          ; <== TRANSMIT TCP_SND_WINDOW bytes
_transmit_not_80:
        movwf       TCPS_LEN
        jump        _tcp_send


_tcp_fin:
        incf        TCPR_SEQ_L0,W           ; tcps_ack1 = B3(tcpr.seq)+1;
        movwf       TCPS_ACK
        incf        TCPS_SEQ,F              ; tcps_seq1++;
        movlw       00010000b               ; tcps_flag = 0x11; // SYN+ACK
                                            ; tcp_online=0;
                                            ; http_state=*;
_set_flag_and_send:
        movwf       TCPS_FLAG
        jump        _tcp_send_empty         ; tcp_send_empty()

_tcp_syn:
    ;;  brset      TCPS_FLAG,B_TCP_ONLINE,_return  ; if( 0==tcp_online ) {
        incf        TCPR_SEQ_L0,W           ; tcps_ack1 = B3(tcpr.seq)+1;
        movwf       TCPS_ACK
        clrf        TCPR_POS                ; tcpr_pos=0;
        clrf        TCPS_SEQ                ; tcps_seq1=ISS=0;
        movlw       10010010b               ; http_state=0; // just connected
        movwf       TCPS_FLAG               ; tcps_flag = 0x12; // SYN+ACK
        call        _tcp_send_empty         ; tcp_send_empty(); // । 0 
        incf        TCPS_SEQ,F              ; tcps_seq1++;
        return                              ; }

if FS_USE_INTERNAL_ROM
;========================================================================
; Internal ROM/EEPROM File (max 512/64 bytes)
;========================================================================
; UNIT0:        ROM             0x0200..0x027F
; UNIT1:        ROM             0x0280..0x02FF
; UNIT2:        ROM             0x0300..0x037F
; UNIT3:        ROM             0x0380..0x03FF
; UNIT4:        EEPROM          0x0000..0x003F
;========================================================================
_file_read:
if FS_USE_INTERNAL_EEPROM
        j0          <HTTP_UNITIDX,2>,_ir_read
_ie_read:
        movf        IP_POS,W
        movwf       EEADR                   ; Address to read
        bsf         RP0
        bsf         EECON1,0                ; EE Read
        bcf         RP0
        movf        EEDATA,W                ; W = EEDATA
        return
endif
_ir_read:
        clrf        PCLATH
        bsf         PCLATH,1
if (FS_INTERNAL_ROM_UNITS >= 2)
        skip0       HTTP_UNITIDX,1
        bsf         PCLATH,0
endif
        movf        IP_POS,W
if (FS_INTERNAL_ROM_UNITS >= 1)
        skip0       HTTP_UNITIDX,0
        iorlw       80h
endif
        movwf       PCL


;========================================================================
; Simple CGI support :)
;========================================================================
_cgi_portb_hi:
        swapf       PORTB,W
        jump        _lo4bit_as_hexdigit
;        skip0       <PORTA,4>
;        retlw       '1'
;        retlw       '0'
_cgi_portb_lo:
        movf        PORTB,W
_lo4bit_as_hexdigit:
        andlw       00001111b
        addlw       -10
        skipnc
        addlw       7
        addlw       58
        return

; .org 800h			;
org 0x200 ,code
;========================================================================
; JavaScript program for PortB control
;========================================================================
        .db 'H','T','T','P','/','1','.','0'
	.db ' ','2','0','0', 13, 10, 13, 10
        ; "Content-Type: %s", "blah-blah-blah"
        ; "Content-Length: %d", blah_blah_blah
        ; "Refresh: %d", blah_blah_blah
        .db '<','t','i','t','l','e','>','P'
	.db 'o','r','t',' ','B','<','/','t'
        .db 'i','t','l','e','>','<','b','o'
	.db 'd','y',' ','o','n','L','o','a'
        .db 'd','=','"','f','o','r','(','v'
	.db 'a','r',' ','i','=','0',';','i'
        .db '<','7',';','i','+','+',')','d'
	.db 'o','c','u','m','e','n','t','.'
        .db 'f','o','r','m','s','[','0',']'
	.db '.','e','l','e','m','e','n','t'
        .db 's','[','i',']','.','c','h','e'
	.db 'c','k','e','d','=','0','x'
        jump        _cgi_portb_hi           ; SIC!
        jump        _cgi_portb_lo
        .db     '&','(','2','<','<','i',')'
	.db '"','>','<','f','o','r','m',' '

        .db 'a','c','t','i','o','n','=','a'
	.db '>','<','s','c','r','i','p','t'
        .db '>','f','o','r','(','v','a','r'
	.db ' ','i','=','1',';','i','<','8'
        .db ';','i','+','+',')','d','o','c'
	.db 'u','m','e','n','t','.','w','r'
        .db 'i','t','e','(','"','B','.','"'
	.db '+','i','+','"','<','I','N','P'
        .db 'U','T',' ','T','Y','P','E','='
	.db 'c','h','e','c','k','b','o','x'
        .db '>','<','B','R','>','"',')','<'
	.db '/','s','c','r','i','p','t','>'
        .db '<','i','n','p','u','t',' ','t'
	.db 'y','p','e','=','s','u','b','m'
        .db 'i','t',' ','v','a','l','u','e'
	.db '=','R','e','a','d','>','<','i'

        .db 'n','p','u','t',' ','t','y','p'
	.db 'e','=','b','u','t','t','o','n'
        .db ' ','v','a','l','u','e','=','W'
	.db 'r','i','t','e',' ','o','n','C'
        .db 'l','i','c','k','=','"','f','o'
	.db 'r','(','v','a','r',' ','v','='
        .db '4','0','0',',','i','=','0',';'
	.db 'i','<','7',';','v','+','=','d'
        .db 'o','c','u','m','e','n','t','.'
	.db 'f','o','r','m','s','[','0',']'
        .db '.','e','l','e','m','e','n','t'
	.db 's','[','i',']','.','c','h','e'
        .db 'c','k','e','d','<','<','+','+'
	.db 'i',')',';','t','o','p','.','l'
        .db 'o','c','a','t','i','o','n','.'
	.db 'h','r','e','f','=','v','"','>'

;========================================================================
; Index Page
;========================================================================
        .db 'H','T','T','P','/','1','.','0'
	.db ' ','2','0','0', 13, 10, 13, 10
        .db '<','a',' ','h','r','e','f','='
	.db 'a','>','P','o','r','t','B',' '
        .db 'c','o','n','t','r','o','l','<'
	.db '/','a','>','<','h','r','>','<'
        .db 'a',' ','h','r','e','f','=','h'
	.db 't','t','p',':','/','/','k','y'
        .db 'l','l','i','k','k','i','.','f'
	.db 'l','u','f','f','.','o','r','g'
        .db '/','h','a','r','d','w','a','r'
	.db 'e','/','>','w','w','w','p','i'
        .db 'c','2',' ','h','o','m','e','p'
        .db 'a','g','e','<','/','a','>',' '
	.db   0,  0,  0,  0,  0,  0,  0,  0
	.db   0,  0,  0,  0,  0,  0,  0,  0

endif

if FS_USE_EXTERNAL_ROM

_file_read:
        retlw       0

endif

end
